#ifndef TABLE_H
#define TABLE_H

#include"get.h"
#include"counter.h"

/* Maximum number of tables per process */
#define TABLE_MAX (64)

/* Maximum number of distinct table types */
#define TABLE_TYPE_MAX (16)

enum table_extent { LOCAL_TABLE, GLOBAL_TABLE  = 2 };
enum table_duplicates { NO_DUPLICATES, DUPLICATES = 4 };
enum table_process { USER_TABLE, KERNEL_TABLE = 8 };
enum table_type { LIST_TABLE = 16, TREE_TABLE = 0, HASH_TABLE = 128 };

union table_base {
	struct binary_base binary;
	struct hash_base hash;
	struct list_base list;
};

typedef int (*compare_fn)(void *, void *);
typedef	int (*hash_fn)(void *, int);

union table_hash {
	hash_fn hash, rehash;
};

struct lock_pool;

struct table {
	struct file *file;
	int type;
	int size[2];
	int count[3];
	union table_base base;
	compare_fn compare;
	union table_hash hash;
	lock_t avail;
	counter_t counter;
	struct lock_pool pool;
};

typedef struct table_ptr {
	struct entry *entry;
	union table_node_ptr {
	} node;
} table_ptr_t;

struct table_op {
	struct table *(*create)(struct table *table);
	void (*destroy)(struct table *table);
	struct table *(*load)(struct table *table, struct file *file);
	int (*store)(struct table *table, struct file *file);
	void *(*lookup)(struct table *table, void *element, struct table_ptr *ptr, int op);
	struct table *(*insert)(struct table *table, void *element, struct table_ptr *ptr);
	struct table *(*delete)(struct table *table, void *element, struct table_ptr *ptr);
	void *(*get)(struct table *table, struct table_ptr *ptr, int get);
	struct table **(*resize)(struct table **table, int size);
};

union table_node {
	struct table table;
	union table_node *next;
};

struct table_pool {
	union table_node base[TABLE_MAX], *avail;
	lock_t free;
};

extern struct table_pool table_pool; /* This is common to all threads */
volatile struct table_op *table_op[TABLE_TYPE_MAX]; /* This is common to all process's */

typedef const struct table * const table_t;

table_t table_create(int type, int size, compare_fn compare, hash_fn hash);
#define table_destroy(table) (counter_active(&(table)->counter) ? 0 :
		table_op[((table)->type / 16) % TABLE_TYPE_MAX]->destroy(table), 1)

void *table_lookup(table_t table, void *element, table_ptr_t *ptr);

table_t table_insert(table_t table, void *element);
table_t table_delete(table_t table, void *element);

void *table_get(table_t table, table_ptr_t *ptr, int get);

table_t *table_resize(table_t *table, int size);

#endif
